@@ -4,13 +4,13 @@
 # found in the LICENSE file.
 
 import logging
-from typing import List, Dict
+from dataclasses import dataclass
+from typing import Dict, List
 
-from . import webassembly, utils
-from .webassembly import OpCode, AtomicOpCode, MemoryOpCode
-from .shared import exit_with_error
+from . import utils, webassembly
 from .settings import settings
-
+from .utils import exit_with_error
+from .webassembly import AtomicOpCode, MemoryOpCode, OpCode
 
 logger = logging.getLogger('extract_metadata')
 
@@ -179,7 +179,7 @@ def find_segment_with_address(module, address):
 
 def data_to_string(data):
   data = data.decode('utf8')
-  # We have at least one test (test/utf8.cpp) that uses a double
+  # We have at least one test (test/core/test_utf8.c) that uses a double
   # backslash in the C++ source code, in order to represent a single backslash.
   # This is because these strings historically were written and read back via
   # JSON and a single slash is interpreted as an escape char there.
@@ -239,19 +239,16 @@ def get_main_reads_params(module, export_map):
   return True
 
 
-def get_named_globals(module, exports):
-  named_globals = {}
-  internal_start_stop_symbols = set(['__start_em_asm', '__stop_em_asm',
-                                     '__start_em_lib_deps', '__stop_em_lib_deps',
-                                     '__em_lib_deps'])
-  internal_prefixes = ('__em_js__', '__em_lib_deps')
+def get_data_exports(module, exports):
+  data_exports = {}
   for export in exports:
     if export.kind == webassembly.ExternType.GLOBAL:
-      if export.name in internal_start_stop_symbols or any(export.name.startswith(p) for p in internal_prefixes):
-        continue
       g = module.get_global(export.index)
-      named_globals[export.name] = str(get_global_value(g))
-  return named_globals
+      # Data symbols (addresses) are exported as immutable Wasm globals.
+      # mutable globals are handled via get_other_exports
+      if not g.mutable:
+        data_exports[export.name] = str(get_global_value(g))
+  return data_exports
 
 
 def get_function_exports(module):
@@ -262,24 +259,46 @@ def get_function_exports(module):
   return rtn
 
 
-def update_metadata(filename, metadata):
-  imports = []
-  invoke_funcs = []
-  with webassembly.Module(filename) as module:
-    for i in module.get_imports():
-      if i.kind == webassembly.ExternType.FUNC:
-        if i.field.startswith('invoke_'):
-          invoke_funcs.append(i.field)
-        else:
-          imports.append(i.field)
-      elif i.kind in (webassembly.ExternType.GLOBAL, webassembly.ExternType.TAG):
+def get_other_exports(module):
+  rtn = []
+  for e in module.get_exports():
+    if e.kind == webassembly.ExternType.FUNC:
+      continue
+    if e.kind == webassembly.ExternType.GLOBAL:
+      g = module.get_global(e.index)
+      # Immutable globals are handled specially.  See get_data_exports
+      if not g.mutable:
+        continue
+    rtn.append(e.name)
+  return rtn
+
+
+def read_module_imports(module, metadata):
+  em_js_funcs = metadata.em_js_funcs
+  types = module.get_types()
+
+  imports = metadata.imports = []
+  invoke_funcs = metadata.invoke_funcs = []
+  em_js_func_types = metadata.em_js_func_types = {}
+
+  for i in module.get_imports():
+    if i.kind == webassembly.ExternType.FUNC:
+      if i.field.startswith('invoke_'):
+        invoke_funcs.append(i.field)
+      else:
+        if i.field in em_js_funcs:
+          em_js_func_types[i.field] = types[i.type]
         imports.append(i.field)
+    elif i.kind in (webassembly.ExternType.GLOBAL, webassembly.ExternType.TAG):
+      imports.append(i.field)
 
+
+def update_metadata(filename, metadata):
+  with webassembly.Module(filename) as module:
     metadata.function_exports = get_function_exports(module)
+    metadata.other_exports = get_other_exports(module)
     metadata.all_exports = [utils.removeprefix(e.name, '__em_js__') for e in module.get_exports()]
-
-  metadata.imports = imports
-  metadata.invokeFuncs = invoke_funcs
+    read_module_imports(module, metadata)
 
 
 def get_string_at(module, address):
@@ -289,31 +308,28 @@ def get_string_at(module, address):
   return data_to_string(data[offset:str_end])
 
 
+@dataclass(init=False)
 class Metadata:
   imports: List[str]
   export: List[str]
-  asmConsts: Dict[int, str]
-  jsDeps: List[str]
-  emJsFuncs: Dict[str, str]
-  emJsFuncTypes: Dict[str, str]
+  em_asm_consts: Dict[int, str]
+  js_deps: List[str]
+  em_js_funcs: Dict[str, str]
+  em_js_func_types: Dict[str, webassembly.FuncType]
   features: List[str]
-  invokeFuncs: List[str]
-  mainReadsParams: bool
-  namedGlobals: List[str]
-
-  def __init__(self):
-    pass
+  invoke_funcs: List[str]
+  main_reads_params: bool
+  data_exports: Dict[str, str]
+  function_exports: Dict[str, webassembly.FuncType]
+  other_exports: List[str]
+  all_exports: List[str]
 
 
 def extract_metadata(filename):
-  import_names = []
-  invoke_funcs = []
   em_js_funcs = {}
-  em_js_func_types = {}
 
   with webassembly.Module(filename) as module:
     exports = module.get_exports()
-    imports = module.get_imports()
 
     export_map = {e.name: e for e in exports}
     for e in exports:
@@ -323,20 +339,8 @@ def extract_metadata(filename):
         string_address = to_unsigned(get_global_value(globl))
         em_js_funcs[name] = get_string_at(module, string_address)
 
-    for i in imports:
-      if i.kind == webassembly.ExternType.FUNC:
-        if i.field.startswith('invoke_'):
-          invoke_funcs.append(i.field)
-        else:
-          if i.field in em_js_funcs:
-            types = module.get_types()
-            em_js_func_types[i.field] = types[i.type]
-          import_names.append(i.field)
-      elif i.kind in (webassembly.ExternType.GLOBAL, webassembly.ExternType.TAG):
-        import_names.append(i.field)
-
-    features = module.parse_features_section()
-    features = ['--enable-' + f[1] for f in features if f[0] == '+']
+    features = module.get_target_features()
+    features = [f'--enable-{feature}' for feature, used in features.items() if used == webassembly.TargetFeaturePrefix.USED]
     features = [f.replace('--enable-atomics', '--enable-threads') for f in features]
     features = [f.replace('--enable-simd128', '--enable-simd') for f in features]
     features = [f.replace('--enable-nontrapping-fptoint', '--enable-nontrapping-float-to-int') for f in features]
@@ -344,17 +348,17 @@ def extract_metadata(filename):
     # If main does not read its parameters, it will just be a stub that
     # calls __original_main (which has no parameters).
     metadata = Metadata()
-    metadata.imports = import_names
     metadata.function_exports = get_function_exports(module)
+    metadata.other_exports = get_other_exports(module)
     metadata.all_exports = [utils.removeprefix(e.name, '__em_js__') for e in exports]
-    metadata.asmConsts = get_section_strings(module, export_map, 'em_asm')
-    metadata.jsDeps = [d for d in get_section_strings(module, export_map, 'em_lib_deps').values() if d]
-    metadata.emJsFuncs = em_js_funcs
-    metadata.emJsFuncTypes = em_js_func_types
+    metadata.em_asm_consts = get_section_strings(module, export_map, 'em_asm')
+    metadata.js_deps = [d for d in get_section_strings(module, export_map, 'em_lib_deps').values() if d]
+    metadata.em_js_funcs = em_js_funcs
     metadata.features = features
-    metadata.invokeFuncs = invoke_funcs
-    metadata.mainReadsParams = get_main_reads_params(module, export_map)
-    metadata.namedGlobals = get_named_globals(module, exports)
+    metadata.main_reads_params = get_main_reads_params(module, export_map)
+    metadata.data_exports = get_data_exports(module, exports)
+
+    read_module_imports(module, metadata)
 
     # print("Metadata parsed: " + pprint.pformat(metadata))
     return metadata
